from __future__ import annotations

from logging import Logger
from typing import Optional

from assertpy.base import BaseMixin
from assertpy.collection import CollectionMixin
from assertpy.contains import ContainsMixin
from assertpy.date import DateMixin
from assertpy.dict import DictMixin
from assertpy.dynamic import DynamicMixin
from assertpy.exception import ExceptionMixin
from assertpy.extracting import ExtractingMixin
from assertpy.file import FileMixin
from assertpy.helpers import HelpersMixin
from assertpy.numeric import NumericMixin
from assertpy.snapshot import SnapshotMixin
from assertpy.string import StringMixin

def assert_that(val: object, description: Optional[str] = "") -> AssertionBuilder: ...

class AssertionBuilder(
    StringMixin,
    SnapshotMixin,
    NumericMixin,
    HelpersMixin,
    FileMixin,
    ExtractingMixin,
    ExceptionMixin,
    DynamicMixin,
    DictMixin,
    DateMixin,
    ContainsMixin,
    CollectionMixin,
    BaseMixin,
    object,
):
    def __init__(
        self,
        val: object,
        description: Optional[str] = "",
        kind: Optional[str] = None,
        expected: Optional[Exception] = None,
        logger: Optional[Logger] = None,
    ) -> None: ...
    def builder(
        self,
        val: object,
        description: Optional[str] = "",
        kind: Optional[str] = None,
        expected: Optional[Exception] = None,
        logger: Optional[Logger] = None,
    ) -> AssertionBuilder: ...
    def error(self, msg: str) -> None: ...



from __future__ import annotations

from datetime import tzinfo
from functools import lru_cache
from typing import Tuple

from arrow import Arrow, get


def __calculate_comparison_date(date_to_compare_to_input: Arrow | str | None, tzinfo: tzinfo) -> Arrow:
    return get(tzinfo=tzinfo) if date_to_compare_to_input is None else get(date_to_compare_to_input)


@lru_cache(maxsize=100)
def convert_inputs_to_dates(
    birthday_input: Arrow | str,
    date_to_compare_to_input: Arrow | str | None = None,
) -> Tuple[Arrow, Arrow]:
    birthday = get(birthday_input)
    comparison_date = __calculate_comparison_date(date_to_compare_to_input, birthday.tzinfo)

    return birthday, comparison_date



from __future__ import annotations

from arrow import Arrow
from delta_calculation import Age, calculate_delta
from input_conversion import convert_inputs_to_dates


def calculate_age(birthday_input: Arrow | str, date_to_compare_to_input: Arrow | str | None = None) -> Age:
    birthday, comparison_date = convert_inputs_to_dates(birthday_input, date_to_compare_to_input)

    return calculate_delta(birthday, comparison_date)



from dataclasses import dataclass
from typing import Tuple

from arrow import Arrow
from dateutil.relativedelta import relativedelta
from typing_extensions import Self

DAYS_IN_WEEK = 7


@dataclass(frozen=True)
class Age:
    years: int
    months: int
    weeks: int
    days: int

    def __neg__(self) -> Self:
        return Age(years=self.years * -1, months=self.months * -1, weeks=self.weeks * -1, days=self.days * -1)


def __calculate_unit(first: Arrow, second: Arrow, unit: str) -> Tuple[int, Arrow]:
    total_of_unit_naive: int = getattr(relativedelta(second.date(), first.date()), unit)

    first_adjusted_naive = first + relativedelta(**{unit: total_of_unit_naive})

    total_of_unit = total_of_unit_naive if first_adjusted_naive <= second else total_of_unit_naive - 1

    return total_of_unit, first + relativedelta(**{unit: total_of_unit})


def __calculate_weeks_and_days(days: int) -> Tuple[int, int]:
    return days // DAYS_IN_WEEK, days % DAYS_IN_WEEK


def __calculate_delta(first: Arrow, second: Arrow) -> Age:
    years, first_adjusted_with_years = __calculate_unit(first, second, "years")
    months, first_adjusted_with_months = __calculate_unit(first_adjusted_with_years, second, "months")
    total_days, _ = __calculate_unit(first_adjusted_with_months, second, "days")
    weeks, days = __calculate_weeks_and_days(total_days)

    return Age(years=years, months=months, weeks=weeks, days=days)


def calculate_delta(first: Arrow, second: Arrow) -> Age:
    if first <= second:
        return __calculate_delta(first, second)

    return -__calculate_delta(second, first)



from __future__ import annotations

from datetime import date
from functools import lru_cache

from arrow import Arrow
from input_conversion import convert_inputs_to_dates


def __calculate_birthday_on_comparison_date_year(birthday: Arrow, comparison_date: Arrow) -> Arrow:
    try:
        return birthday.replace(year=comparison_date.year)
    except ValueError:
        return birthday.replace(year=comparison_date.year, day=birthday.day - 1)


@lru_cache(maxsize=100)
def calculate_next_birthday(birthday_input: Arrow | str, date_to_compare_to_input: Arrow | str | None = None) -> date:
    birthday, comparison_date = convert_inputs_to_dates(birthday_input, date_to_compare_to_input)

    birthday_on_comparison_date_year = __calculate_birthday_on_comparison_date_year(birthday, comparison_date)

    return (
        birthday_on_comparison_date_year.replace(year=birthday_on_comparison_date_year.year + 1)
        if birthday_on_comparison_date_year <= comparison_date
        else birthday_on_comparison_date_year
    ).date()



from __future__ import annotations

from datetime import date, timedelta
from typing import Optional, TypeAlias, Union, overload

from typing_extensions import Self

RelativeDeltaCompat: TypeAlias = Union[timedelta, date]

class relativedelta(object):
    years: int
    months: int
    days: int
    leapdays: int
    hours: int
    minutes: int
    seconds: int
    microseconds: int
    year: Optional[int]
    month: Optional[int]
    day: Optional[int]
    weekday: Optional[int]
    hour: Optional[int]
    minute: Optional[int]
    second: Optional[int]
    microsecond: Optional[int]
    _has_time: int

    def __init__(
        self,
        dt1: Optional[date] = None,
        dt2: Optional[date] = None,
        years: int = 0,
        months: int = 0,
        days: int = 0,
        leapdays: int = 0,
        weeks: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: int = 0,
        microseconds: int = 0,
        year: Optional[int] = None,
        month: Optional[int] = None,
        day: Optional[int] = None,
        weekday: Optional[int] = None,
        yearday: Optional[int] = None,
        nlyearday: Optional[int] = None,
        hour: Optional[int] = None,
        minute: Optional[int] = None,
        second: Optional[int] = None,
        microsecond: Optional[int] = None,
    ) -> None: ...
    def _fix(self) -> None: ...
    @overload
    def weeks(self) -> int: ...
    @weeks.setter
    def weeks(self, value: int) -> None: ...
    def _set_months(self, months: int) -> None: ...
    def normalized(self) -> Self: ...
    def __add__(self, other: Union[Self, RelativeDeltaCompat]) -> Self: ...
    def __radd__(self, other: Union[Self, RelativeDeltaCompat]) -> Self: ...
    def __rsub__(self, other: Union[Self, RelativeDeltaCompat]) -> Self: ...
    def __sub__(self, other: Union[Self, RelativeDeltaCompat]) -> Self: ...
    def __abs__(self) -> Self: ...
    def __neg__(self) -> Self: ...
    def __bool__(self) -> bool: ...
    # Compatibility with Python 2.x
    __nonzero__ = __bool__

    def __mul__(self, other: Union[Self, RelativeDeltaCompat]) -> Self: ...
    __rmul__ = __mul__
    def __eq__(self, other: Union[Self, RelativeDeltaCompat]) -> Self: ...
    def __hash__(self) -> int: ...
    def __ne__(self, other: Union[Self, RelativeDeltaCompat]) -> bool: ...
    def __div__(self, other: Union[Self, RelativeDeltaCompat]) -> Self: ...
    __truediv__ = __div__

    def __repr__(self) -> str: ...

def _sign(x: float) -> int: ...
